#pragma once

#include <vector>
using namespace std;

class SDSceneNode;
class SDTreeNode;

#define MAXUNSIGNEDINTEGERSIZE 4294967294

// This class simplifies determining whether a node has been rendered or not.
// No need to clear any flags in the instances, the static clear will invalidate
// all trues. After the program has been run long enough. i.e after
// MAXUNSIGNEDINTEGERSIZE resetAll's some false positives may be reported.
// It is very unlikely, since a node must not be rendered in the time
// between MAXUNSIGNEDINTEGERSIZE frames. At 60 fps this will take 19000 hours to
// accomplish. If it does happen, that node will not be rendered that frame. But will
// be rendered the next. Is it worth the extra code to consider this case? I think not
class SDSceneNodeState
{
protected:
	static unsigned int m_i;	
	unsigned int		m_value;

public:
	static bool resetAll() 
	{
		m_i++;
		if ( m_i > MAXUNSIGNEDINTEGERSIZE )
			m_i = 1;
	}

	static void	setTrue( SDSceneNodeState s )
	{
		// Set to the current rendered state
		s.m_value = m_i;
	}

	static bool isTrue( SDSceneNodeState s )
	{
		if( m_i == s.m_value )
			return true;
		else
			return false;
	}

};

// A list of all the root scene nodes used by a visualization
class SDSceneNodeInfo
{
protected:
	// The actual node object
	SDSceneNode*		m_pNode;

	// List of any tree nodes that contain this node
	vector<SDTreeNode*>	m_treeLocations;

	// Maximum number of references to this object allowed
	int					m_maxSplitDepthAllowed;

	// List of any display lists that contain this node

	// Whether this node was added to a display list already
	SDSceneNodeState	m_isDrawn;

public:
	SDSceneNodeInfo() { m_maxSplitDepthAllowed = 2; }
	virtual ~SDSceneNodeInfo();

	void	setNodeObject( SDSceneNode* pNode)			{ m_pNode = pNode; }
	void	addTreeLocation(SDTreeNode* t)				{ if( t != NULL) m_treeLocations.push_back(t); }
	void	removeTreeLocation(SDTreeNode* t);

	int				getNumberLocations()		{ return (int)m_treeLocations.size(); }
	SDTreeNode*		getLocation( int i )		{ return m_treeLocations[i]; }
	int				getMaxSplitDepthAllowed()	{ return m_maxSplitDepthAllowed; }
	SDSceneNode*	getNodeObject()				{ return m_pNode; }
	bool			isDrawn()					{ return SDSceneNodeState::isTrue(m_isDrawn); }

};


class SDNodeList
{
protected:

	// The list of scene nodes
	vector<SDSceneNodeInfo*>	m_nodes;

public:
	SDNodeList(void);
	virtual ~SDNodeList(void);

	virtual int					getNodeListSize()			{ return (int)m_nodes.size(); }
	virtual SDSceneNodeInfo*	getNodeInfo(int index)		{ return m_nodes[index]; }
	virtual int					addNode(SDSceneNode* pNode);

};
