#include "sd.h"

SDOctTreeNode::SDOctTreeNode(SDNodeList* pList, SDOctTreeNode* pParent, SD3DBox bounds, int depth) : SDTreeNode(pList, depth)
{

	// Log creating this octtree node
	Point c;
	Point d;
	bounds.get(&c, &d);
	if( depth <= 1)
		::log("Created octtree at depth %d: Centre: (%.3f, %.3f, %.3f) Dimensions: (%.3f, %.3f, %.3f)", depth, c.x, c.y, c.z, d.x, d.y, d.z);

	// Set the parent
	m_pParent = pParent;

	// Set the bounds
	m_bounds = bounds;

	//Create children if the bounding box is big enough
	Point w;
	m_bounds.getDimensions(&w);
	w = w/2;
	if( w.x >= OCTTREE_MINNODESIZE || w.y >= OCTTREE_MINNODESIZE || w.z >= OCTTREE_MINNODESIZE )
	{
		SD3DBox b;
		Point c, d;
		int i;
		for(i=0; i<8; i++)
		{
			calculateChildBox(&c, &d, i);
			b.set(c, d);
			m_pChildren[i] = new SDOctTreeNode(pList, this, b, depth + 1);
		}
	}
	else
	{
		int i;
		for(i=0; i<8; i++)
			m_pChildren[i] = NULL;
	}
	
}

SDOctTreeNode::~SDOctTreeNode(void)
{

	Point c;
	Point d;
	m_bounds.get(&c, &d);
	if( m_depth <= 1)
		::log("Deleting octtree at depth %d: Centre: (%.3f, %.3f, %.3f) Dimensions: (%.3f, %.3f, %.3f)", m_depth, c.x, c.y, c.z, d.x, d.y, d.z);


	// Delete the children
	int i;
	for(i=0; i<8; i++)
	{
		if( m_pChildren[i])
		{
			delete m_pChildren[i];
			m_pChildren[i] = NULL;
		}
	}

}

// called when a node moves position so the tree can update its location
void SDOctTreeNode::nodeMoved( int nodeIndex )
{
}

void SDOctTreeNode::calculateChildBox(Point* c, Point* d, int child)
{
	// Get this node's bounds
	m_bounds.get(c, d);

	// Divide the bound dimensions in 2
	(*d) = (*d)/2;
	// Calculate the correct centre
	if( child & PLUSZPLANE )	
		c->z = c->z + d->z/2;
	else
		c->z = c->z - d->z/2;
	// Calculate the correct centre
	if( child & PLUSYPLANE )	
		c->y = c->y + d->y/2;
	else
		c->y = c->y - d->y/2;
	// Calculate the correct centre
	if( child & PLUSXPLANE )	
		c->x = c->x + d->x/2;
	else
		c->x = c->x - d->x/2;

}

// Adds a node to the tree.
bool SDOctTreeNode::addNode(int nodeIndex)
{
	return addNode(nodeIndex, 0);
}

// Adds a node to the tree and ensures it isn't split past the maximum split depth
bool SDOctTreeNode::addNode(int nodeIndex, int splitDepth)
{

	SD3DBox*			pChildBounds = NULL;
	SDSceneNodeInfo*	pNodeInfo = m_pNodes->getNodeInfo(nodeIndex);
	SDSceneNode*		pNode = pNodeInfo->getNodeObject();
	int i;

	if( pNode == NULL )
	{
		::log("Unable to retrieve node object from node list index");
		return false;
	}

	// Check if there are no children
	if( m_pChildren[0] == NULL)
	{
		Point c, d;
		m_bounds.get(&c, &d);
		::log("Adding node to tree node with bounding box: Centre(%.3f,%.3f,%.3f) Dim(%.3f,%.3f,%.3f)", c.x, c.y, c.z, d.x, d.y, d.z);
		m_pNodeRefList->push_back(nodeIndex);
		pNodeInfo->addTreeLocation(this);
		return true;
	}

	// The results of the fits tests
	bool		intersections[8];
	ZeroMemory( intersections, 8*sizeof(bool));
	Point		c;
	int			iInt;
	m_bounds.getLocation(&c);

	//Check which octants this node goes in
	iInt = pNode->whichOctants(&c, intersections);

	// If there is a unique octant that this node goes in.. then add it to that one. Dont increment the split count since we didn't split it
	if( iInt != -1 )
	{
		return m_pChildren[iInt]->addNode(nodeIndex, splitDepth);
	}
	else
	{
		// Check if the object has reached the maximum number of refereneces allowed
		if( splitDepth >= pNodeInfo->getMaxSplitDepthAllowed() )
		{
			Point c, d;
			m_bounds.get(&c, &d);
			::log("Adding node to tree node with bounding box: Centre(%.3f,%.3f,%.3f) Dim(%.3f,%.3f,%.3f)", c.x, c.y, c.z, d.x, d.y, d.z);
			m_pNodeRefList->push_back(nodeIndex);
			pNodeInfo->addTreeLocation(this);
		}
		else
		{
			// Add to the children which the node intersected
			for( i = 0; i < 8; i++)
			{
				if(intersections[i])
					m_pChildren[i]->addNode(nodeIndex, splitDepth+1);
			}
		}
	}	

	return true;
}

// Removes a node from this tree
bool SDOctTreeNode::removeNode( int nodeIndex )
{
	if( nodeIndex > (int)m_pNodeRefList->size() )
		return false;

	SDSceneNodeInfo* pNodeInfo = m_pNodes->getNodeInfo(nodeIndex);
		
#if 0
	// Test if the node has a link to this tree
	int i;
	SDSceneNodeTree* l = NULL;
	int size = pNodeInfo->getNumberLocations();
	for(i = 0; i < size; i++)
	{
		l = pNodeInfo->getLocation(i);
		if(l == this)
			break;
	}

	// Check if this tree was not found in the list
	if(i > size)
	{
		::log("SDTreeNode::RemoveNode() Unable to find a reference to the tree in the node info");
		return false;
	}
#endif	

	// Erase the node from this list
	list<int>::iterator i = m_pNodeRefList->begin();
	while(i != m_pNodeRefList->end() )
	{
		if( *i = nodeIndex )
		{
			m_pNodeRefList->erase(i);
			return true;
		}
		i++;
	}
	
	return false;

}