#include <fstream>
using namespace std;

#include "vcsimulationlink.h"
#include "sd.h"
#include "sl.h"

SLLogFile::SLLogFile(VCSimulationLink* pLink, string logFile) : SLBase(pLink)
{
	m_logFile		= logFile;	
}

SLLogFile::~SLLogFile(void)
{
}

// Divides a string into tokens based on separators in the deliminator string
void SLLogFile::tokenize(vector<string>* tokens, string line, string delims)
{

	// The outlines of the current token
	unsigned int left = 0;
	unsigned int right = 0;

	while( right < line.length() )
	{
		// Check if right is now at a delimiter
		if( delims.find(line.at(right)) != string::npos )
		{
			// Add the token 
			if( left != right )
				tokens->push_back(line.substr(left, right-left));

			// Set the left and right back to being equal and start searching for the next token after this one
			right++;
			left = right;
		}
		else
		{
			// Just keep looking for a delimiter
			right++;
		}
	}

	// Grab the last token if there is one
	if( right != left && right > 0)
	{
		tokens->push_back(line.substr(left, right - left));
	}	

}

// Sets the model and event values from the token list
void SLLogFile::createVisualizationObjects(SDDevsModel* msrc, SDDevsModel* mdest, SDEvent* e, vector<string>& tokens)
{

	// Set the type
	if( tokens[TOKEN_TYPE] == "X" )
		e->m_type = EXTERNAL;
	else
		e->m_type = OUTPUT;

	// Set the time
	e->m_time.setTime(tokens[TOKEN_TIME]);

	// Set the source
	// Trim the object number at the end.. no need for that
	size_t right	= tokens[TOKEN_SOURCE].find_last_of("(");
	e->m_source	= tokens[TOKEN_SOURCE].substr(0, right);

	// Set the port
	e->m_port	= tokens[TOKEN_PORT];

	// Set the value
	sscanf(tokens[TOKEN_VALUE].c_str(), "%lf", &e->m_value);

	// Set the destination
	// Trim the object number at the end.. no need for that
	right	= tokens[TOKEN_DESTINATION].find_last_of("(");
	e->m_destination = tokens[TOKEN_DESTINATION].substr(0, right);

	// Set up the model based on the event
	//---------

	// Add the source as a model
	msrc->m_name = e->m_source;
	// If its an output event... the port is associated
	// with the sender, and is an output port
	if(e->m_type == OUTPUT)
		msrc->m_outputPorts.push_back(e->m_port);
	msrc->m_pSceneNode = NULL;
	msrc->m_state = 0.0;

	// Add the destination as a model
	mdest->m_name = e->m_destination;
	// If its an external event... the port is associated
	// with the reciever, and is an input port
	if(e->m_type == EXTERNAL)
		mdest->m_inputPorts.push_back(e->m_port);
	mdest->m_pSceneNode = NULL;
	mdest->m_state = 0.0;

}

// Parses a line from a log file to produce an event
bool SLLogFile::parse(char* line)
{
	//*************************************************************************************************************
	// The format for the log file is so:
	// For the types that concern the visualization( type = 'Y' and type = 'X' )	
	// Mensage $(char type) / $(SDTime time) / $(string source)[$(int x, int y, int z, ...)](#) / $(string port_name) / $(double value) para $(string destination)[($(int x) , $(int y), $(int z), ...)](#)
	
	// # means any integer
	// $(type name) = variable of type 'type'
	
	// Check if the Message type is X or Y
	vector<string>	tokens;
	tokenize(&tokens, string(line), string("/ "));

	if( tokens.size() !=  8 )
	{
		//::log("Ignoring line with non X or Y format type.");
		return true;
	}

	if( tokens[1] == "X" || tokens[1] == "Y" )
	{
		::log("Adding event");
		::log("%s", line);
	}
	else
	{
		::log("%s", tokens[1].c_str());
		::log("Ignoring non external or output message");
		return true;
	}

	//****************************
	// (Todo) NOTE: 
	// Add the devs models implied in this log file to the visualization
	// if they haven't been added.
	// Treat cell-devs models differently since they have an implicit parent child relationship defined.
	//****************************

	// Create a model for the reciever and sender in this event
	SDDevsModel* msrc	= new SDDevsModel();
	SDDevsModel* mdest	= new SDDevsModel();
	// Create the event associated with this token
	SDEvent* e = new SDEvent();
	createVisualizationObjects(msrc, mdest, e, tokens);

	// Check if the models are new
	if( !m_pLinkControl->notifyDEVSModel(msrc) )
		delete msrc;
	if( !m_pLinkControl->notifyDEVSModel(mdest) )
		delete mdest;

	// Check if the event is new
	if( !m_pLinkControl->notifyEvent(e) )
	{
		::log("***Event already in event list***");
		// The event was not stored
		delete e;
	}


	return true;
}

// Loads a log file
bool SLLogFile::connect()
{

	// Ensure this file exists and can be loaded
	ifstream* pLogFile = new ifstream();

	if(!pLogFile)
	{
		::log("Unable to create ifstream object.");
		return false;
	}

	// Open the file
	pLogFile->open(m_logFile.c_str());
	if( !pLogFile->good() )
	{
		::log("Unable to open log file for import");
		return false;
	}
	
	// Parse the lines of the log file.
	//------------

	// Each Line is an event and will have a source and destination devs model

	// Store the line count
	int i = 0;
	while( !pLogFile->eof() )
	{
		// Get the line with maximum 512 characters
		char buffer[512];
		pLogFile->getline(buffer, 512);

		// Parse the line
		if(!parse(buffer))
		{
			::log("Unable to parse line %d of log file: %s", i, m_logFile.c_str() );
			return false;
		}

		//Increment the line count
		i++;
	}

	return true;

}
