#pragma once
#define DEFAULT_MAX_LOG_SIZE 300
#define CONSOLE_SPEED 0.5		// Time it takes in seconds for the console to 
								// appear / disapear

enum ConsoleState {CS_VISIBLE, CS_HIDDEN, CS_WORKING, CS_SHOWING, CS_HIDING};
enum CommandResult {
	CR_NOT_DEFINED = 0x0,
	CR_ADD_TO_HISTORY = 0x40,
	CR_SUCCESS_NO_HISTORY = 0x1,
	CR_SUCCESS = CR_SUCCESS_NO_HISTORY | CR_ADD_TO_HISTORY, 

	CR_ERROR = 0x80,
	CR_USAGE_ERROR = CR_ERROR | 0x2
};

typedef vector<string> ConsoleLog;
typedef ConsoleLog::iterator ConsoleLogIterator;

typedef vector<string> CommandHistory;
typedef CommandHistory::iterator CommandHistoryIterator;

typedef CommandResult (CALLBACK* CommandProc)(void*, string, vector<string>);

struct CommandEventInfo {
	void* m_object;
	CommandProc m_callback;

	CommandEventInfo(void* object, CommandProc callback) {
		m_object = object;
		m_callback = callback;
	};
};

typedef vector<CommandEventInfo*> CommandCallbackList;
typedef CommandCallbackList::iterator CommandCallbackIterator;


class Console :
	public UserInterface
{
private:
	static Console s_console;

	ConsoleState m_state;
	
	ConsoleLog m_log;
	
	CommandHistory m_history;
	int m_historyIndex;

	Font *m_consoleFont;

	EditBox m_editBox;
	Scrollbar m_scrollbar;

	Texture *m_texture;

	float m_movement;			// Used when showing/hiding

	int m_maxLogSize;
	int m_maxVisibleLines;
	int m_firstVisibleLine;

	RGBA m_logColor;
	RGBA m_commandColor;

private:
	static CommandResult CALLBACK ConsoleCommandWatcher(void* object, string command, vector<string> parameters);
	static void CALLBACK OnScroll(Scrollbar *scrollbar, float value);
	static void CALLBACK OnEngineVariableChange(void * object, string variable, string newValue);
public:
	static Console *Get() {return &s_console;};	

	// IInputConsumer Methods
	virtual bool processKeyUp(unsigned char key, int x, int y) {return false;};
	virtual bool processKeyDown(unsigned char key, int x, int y);
	
	virtual bool processSpecialKeyUp(int key, int x, int y){return false;};
	virtual bool processSpecialKeyDown(int key, int x, int y);

	virtual bool processMouseMove(int x, int y){return false;};
	virtual bool processMouseClick(int button, int state, int x, int y){return false;};

	// ITickable Methods
	virtual void tick();
	// IDrawable Methods
	virtual void draw();

	// IDisposable Methods
	virtual void dispose(){return;};
	// UserInterface Methods
	virtual void showCursor(bool show) {};

	ConsoleState getState() {return m_state;};
	
	void show();
	void hide();
	void toggleVisible();
	
	void writeline(string text);
	void write(string text);
	void write(char* format, ...);

	void clear() {m_log.clear(); m_scrollbar.setScrollValues(0, 0, 0);};
	void clearCommandHistory() {m_history.clear();};

	void processCommand(string command, bool echo = true);
	void addCommandCallback(void* object, CommandProc callback);

	void initialize();

private:
	void setupControls();
	void addToHistory(string command);

public:
	Console(void);

	~Console(void) {
		m_consoleFont->dispose();		
		delete m_consoleFont;

		m_log.clear();
		m_history.clear();

		m_texture->unload();
		delete m_texture;
	};

private:
	
	class CommandParser {
	private:
		CommandCallbackList m_callbacks;

		string extractString(string search, int &position);
		CommandResult sendNotifications(string command, vector<string> parameters);

	public:
		bool parse(string commandline);	
		void addCallback(void* object, CommandProc callback);

	public:
		CommandParser(void) {};
		~CommandParser(void) {
			m_callbacks.clear();
		};
	} m_parser;
	
};
