#include "includes.h"

/*static*/ GLExtensions* GLExtensions::s_pGLExtensions = NULL;

// Extention functions
//-----------

// Multitexture functions
PFNGLACTIVETEXTUREARBPROC		pglActiveTextureARB			= NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC pglClientActiveTextureARB	= NULL;
PFNGLMULTITEXCOORD3DARBPROC		pglMultiTexCoord3fARB		= NULL;

// Vertex Buffer functions
PFNGLBINDBUFFERARBPROC			pglBindBufferARB	= NULL;
PFNGLGENBUFFERSARBPROC			pglGenBuffersARB	= NULL;
PFNGLBUFFERDATAARBPROC			pglBufferDataARB	= NULL;
PFNGLDELETEBUFFERSARBPROC		pglDeleteBuffersARB = NULL;

// Vertex and Fragment program functions
PFNGLPROGRAMSTRINGARBPROC		pglProgramStringARB		= NULL;
PFNGLBINDPROGRAMARBPROC			pglBindProgramARB		= NULL;
PFNGLDELETEPROGRAMSARBPROC		pglDeleteProgramsARB	= NULL;
PFNGLGENPROGRAMSARBPROC			pglGenProgramsARB		= NULL;
PFNGLPROGRAMLOCALPARAMETER4FVARBPROC	pglProgramLocalParameter4fvARB	= NULL;

// Function used to get the extension procedures
inline void *getAddress (const char *name) {
	#ifdef _WIN32
		return (void *) wglGetProcAddress (name);
	#else
		return (void *) glXGetProcAddress ((const GLubyte *) name);
	#endif /* _WIN32 */
}

// Tests if an extension is supported
int isExtensionSupported(const char *extension)
{
  const GLubyte *extensions = NULL;
  const GLubyte *start;
  GLubyte *where, *terminator;

  /* Extension names should not have spaces. */
  where = (GLubyte *) strchr(extension, ' ');
  if (where || *extension == '\0')
    return 0;
  extensions = glGetString(GL_EXTENSIONS);
  /* It takes a bit of care to be fool-proof about parsing the
     OpenGL extensions string. Don't be fooled by sub-strings,
     etc. */
  start = extensions;
  for (;;) {
    where = (GLubyte *) strstr((const char *) start, extension);
    if (!where)
      break;
    terminator = where + strlen(extension);
    if (where == start || *(where - 1) == ' ')
      if (*terminator == ' ' || *terminator == '\0')
        return 1;
    start = terminator;
  }
  return 0;
}

GLExtensions::GLExtensions()
{

	// Detect and get the extensions
	m_multiTexture = false;
	m_vertexBuffers = false;
	m_vertexPrograms = false;
	m_fragmentPrograms = false;

	// Check for multitexture support
	if (isExtensionSupported("GL_ARB_multitexture"))
	{
		pglActiveTextureARB			= (PFNGLACTIVETEXTUREARBPROC) getAddress("glActiveTextureARB");
		pglClientActiveTextureARB	= (PFNGLCLIENTACTIVETEXTUREARBPROC) getAddress("glClientActiveTextureARB");
		pglMultiTexCoord3fARB		= (PFNGLMULTITEXCOORD3DARBPROC) getAddress("glMultiTexCoord3fARB");
		::log("GL_ARB_multitexture extension found");
		m_multiTexture = true;
	}

	// Check for vertex buffers
	if(isExtensionSupported("GL_ARB_vertex_buffer_object"))
	{
		pglBindBufferARB	= (PFNGLBINDBUFFERARBPROC) getAddress("glBindBufferARB");
		pglGenBuffersARB	= (PFNGLGENBUFFERSARBPROC) getAddress("glGenBuffersARB");
		pglBufferDataARB	= (PFNGLBUFFERDATAARBPROC) getAddress("glBufferDataARB");
		pglDeleteBuffersARB = (PFNGLDELETEBUFFERSARBPROC) getAddress("glDeleteBuffersARB");
		::log("GL_ARB_vertex_buffer_object extension found");
		m_vertexBuffers = true;
	}

	// Check for vertex programs
	if(isExtensionSupported("GL_ARB_vertex_program"))
	{
		pglProgramStringARB			= (PFNGLPROGRAMSTRINGARBPROC) getAddress("glProgramStringARB");
		pglBindProgramARB			= (PFNGLBINDPROGRAMARBPROC) getAddress("glBindProgramARB");
		pglDeleteProgramsARB		= (PFNGLDELETEPROGRAMSARBPROC) getAddress("glDeleteProgramsARB");
		pglGenProgramsARB			= (PFNGLGENPROGRAMSARBPROC) getAddress("glGenProgramsARB");
		pglProgramLocalParameter4fvARB	= (PFNGLPROGRAMLOCALPARAMETER4FVARBPROC) getAddress("glProgramLocalParameter4fvARB");
		::log("GL_ARB_vertex_program extension found");
		m_vertexPrograms = true;
	}

	// Check for fragment programs.. if they are available then the vertex programs better have been enabled
	if(m_vertexPrograms && isExtensionSupported("GL_ARB_fragment_program"))
	{
		::log("GL_ARB_fragment_program extension found");
		m_fragmentPrograms = true;
	}

}

GLExtensions::~GLExtensions()
{
}

// Multitexture extension support
void GLExtensions::glActiveTextureARB(GLenum texture) { if(pglActiveTextureARB) { return pglActiveTextureARB(texture); } else { ::log("glActiveTextureARB called without corresponding OpenGL extension support"); } }
void GLExtensions::glClientActiveTextureARB(GLenum target) { if(pglClientActiveTextureARB) { return pglClientActiveTextureARB(target); } else { ::log("glClientActiveTextureARB called without corresponding OpenGL extension support"); } }
void GLExtensions::glMultiTexCoord3fARB(GLenum texture, GLfloat s, GLfloat t, GLfloat r) { if(pglMultiTexCoord3fARB) { return pglMultiTexCoord3fARB(texture, s, t, r); } else { ::log("glMultiTexCoord3fARB called without corresponding OpenGL extension support"); } }

// Vertex buffer extension support
void GLExtensions::glBindBufferARB(GLenum target, GLuint buffer) { if(pglBindBufferARB) { return pglBindBufferARB(target, buffer); } else { ::log("glBindBufferARB called without corresponding OpenGL extension support"); } }
void GLExtensions::glGenBuffersARB(GLsizei n, GLuint *buffers) { if(pglGenBuffersARB) { return pglGenBuffersARB(n, buffers); } else { ::log("glGenBuffersARB called without corresponding OpenGL extension support"); } }
void GLExtensions::glBufferDataARB(GLenum target, GLsizeiptrARB size, const GLvoid *data, GLenum usage) { if(pglBufferDataARB) { return pglBufferDataARB(target, size, data, usage); } else { ::log("glBufferDataARB called without corresponding OpenGL extension support"); } }
void GLExtensions::glDeleteBuffersARB(GLsizei n, GLuint *buffers) { if(pglDeleteBuffersARB) { return pglDeleteBuffersARB(n, buffers); } else { ::log("glDeleteBuffersARB called without corresponding OpenGL extension support"); } }

// Fragment and Vertex Program support
void GLExtensions::glProgramStringARB(GLenum target, GLenum format, GLsizei len, const GLvoid *string) { if(pglProgramStringARB) { return pglProgramStringARB(target, format, len, string); } else { ::log("glProgramStringARB called without corresponding OpenGL extension support"); } }
void GLExtensions::glBindProgramARB(GLenum target, GLuint program) { if(pglBindProgramARB) { return pglBindProgramARB(target, program); } else { ::log("glBindProgramARB called without corresponding OpenGL extension support"); } }
void GLExtensions::glDeleteProgramsARB(GLsizei n, const GLuint *programs) { if(pglDeleteProgramsARB) { return pglDeleteProgramsARB(n, programs); } else { ::log("glDeleteProgramsARB called without corresponding OpenGL extension support"); } }
void GLExtensions::glGenProgramsARB(GLsizei n, GLuint *programs) { if(pglGenProgramsARB) { return pglGenProgramsARB(n, programs); } else { ::log("glGenProgramsARB called without corresponding OpenGL extension support"); } }
void GLExtensions::glProgramLocalParameter4fvARB(GLenum target, GLuint index, const GLfloat* params) { if(pglProgramLocalParameter4fvARB) { return pglProgramLocalParameter4fvARB(target, index, params); } else { ::log("glProgramLocalParameter4fvARB called without corresponding OpenGL extension support"); } }

