#include "includes.h"

//*****************************************************************************************//
//                          Private Packing/Unpacking Facilities                           //
//*****************************************************************************************//

#define redMask 0X000000FF
#define greenMask 0X0000FF00
#define blueMask 0X00FF0000
#define alphaMask 0XFF000000

#define redShift 0
#define greenShift 8
#define blueShift 16
#define alphaShift 24

inline long pack (const long red, const long green, const long blue, const long alpha) {
	return 
		((red << redShift) & redMask) | 
		((green << greenShift) & greenMask) | 
		((blue << blueShift) & blueMask)| 
		((alpha << alphaShift) & alphaMask);
}

inline long unpackRed (const long pixel) {
	return (pixel >> redShift) & 0X000000FF;
}

inline long unpackGreen (const long pixel) {
	return (pixel >> greenShift) & 0X000000FF;
}

inline long unpackBlue (const long pixel) {
	return (pixel >> blueShift) & 0X000000FF;
}

inline long unpackAlpha (const long pixel) {
	return (pixel >> alphaShift) & 0X000000FF;
}

inline bool isPowerOf2 (long value) {
	return (value & ~(value - 1)) == value;
}	

//------------------------------------------------------------------------------------------

HeightMap::HeightMap()
{
	this->m_pHeights	= NULL;
	this->m_length		= 0;
	this->m_width		= 0;
}

HeightMap::~HeightMap()
{
	if(this->m_pHeights != NULL)
	{
		delete [] m_pHeights;
		m_pHeights = NULL;
	}
}


//Private utility for loadHeightMap
inline char *fileSuffix (const char *fileName) {
	//Returns a pointer to an ALL UPPERCASE string in a static area containing the 3 CHARACTER SUFFIX SUCH AS
	//"BMP" OR "TGA". Don't call twice on 2 different file names and then process (call once, process, call 
	//again, process, ...).
	long size = (long)strlen (fileName); static char uppercase [5];
	for (char *s = (char *) &fileName [size - 4], *d = &uppercase [0]; *s; s++, d++) {
		*d = toupper (*s);
	}
	uppercase [4] = '\0'; //Make sure the file extension ends with a NULL
	return uppercase;
}

// Loads a heightmap from a bitmap, targa or other file
bool HeightMap::loadHeightMap(char* fileName)
{

	// Get the suffix from the file name to determine the type of file we are loading.
	// Currently we can load from bitmaps or targas.
	char *suffix = fileSuffix (fileName);

	if (strcmp (suffix, ".BMP") == 0) 
		return readBMPHeightMap (fileName);
	if (strcmp (suffix, ".TGA") == 0) 
		return readTGAHeightMap (fileName);

	halt ("\nUnknown heightmap type requested for \"%s\"...", fileName);	
	return false;
}

bool HeightMap::readBMPHeightMap (char *fileName) {
	//Read a heightmap and fill in all attributes; returns NULL if unsuccessful.
	//Creates and fills the unsigned char array of heightmap values

	//Read the bitmap in using the Win32 API.
    HBITMAP bitmapHandle = (HBITMAP) LoadImage (NULL, fileName, 
    	IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE | LR_CREATEDIBSECTION);	
	if (bitmapHandle == NULL) {log ("\nFile \"%s\" not found...", fileName); return NULL;}

    //Find out how big it is.
	BITMAP bitmap; GetObject (bitmapHandle, sizeof (bitmap), &bitmap);
	long width = bitmap.bmWidth;	
	long byteWidth = bitmap.bmWidthBytes; 
	long height = bitmap.bmHeight;
	long pixelSize = bitmap.bmBitsPixel;
	long planes = bitmap.bmPlanes;
	char *bits = (char *) bitmap.bmBits;

	if (!isPowerOf2 (width) || !isPowerOf2 (height)) {
		//The heightmap need not be a power of 2 but it is useful for quad tree visibility so lets enforce it for now.
		prompt ("\nBitmap \"%s\" is not a power of 2; width %d, height %d.", fileName, width, height);
		log ("\nBitmap \"%s\" is not a power of 2; width %d, height %d.", fileName, width, height);
	}

	//Allocate space the heightmap.
	m_pHeights = new unsigned char [width * height];

	// Set the heightmap dimensions
	m_length = height;
	m_width = width;

	if (m_pHeights == NULL) {DeleteObject (bitmapHandle); return false;}

	//Prepare to transfer the bits.
	unsigned char *destination = m_pHeights; //REVERSES: destination += height * width; //past the end
    BYTE *source = (BYTE *) bits; //at the beginning

	//Move the bits from the bitmap handle to the heightmap.	 
	long red, green, blue; long alpha = 255; //0=>transparent, 255=>opaque
	if (pixelSize == 24) {
		//Successively read 3 RGB bytes at a time, write RGBA bytes. 
    	for	(long j = height; j > 0; j--) {
			//REVERSES: destination -= width; 
			unsigned char *to = destination; BYTE *from = source;
    		for (int i = 0; i < width; i++) { 
				blue = *from++;	green = *from++; red = *from++;
				*to++ = (unsigned char)red;
    		}   
			destination += width; 											   
			source += byteWidth;
		}

/*		for(int j=0; j<height; j++)
		{
			for(int i=0; i<width; i++)
			{
				unsigned char stuff = m_pHeights[j*width + i];
				int a = stuff;
			}
		}*/
	} else if (pixelSize == 8) { 
		//Obtains the palette information from the handle.
		typedef PALETTEENTRY PaletteEntries [256];
		HDC deviceContextHandle = CreateCompatibleDC (NULL);
		SelectObject (deviceContextHandle, bitmapHandle);
		PaletteEntries paletteEntries; 
		ZeroMemory (&paletteEntries, sizeof (PaletteEntries));
		GetDIBColorTable (deviceContextHandle, 0, 256, (RGBQUAD *) &paletteEntries);
		DeleteDC (deviceContextHandle);	
		long packedPaletteEntries [256];

		//Pack it into a nice RGBA array.
		long *palette = (long *) &paletteEntries; 
		for (int i = 0; i < 256; i++) {
			//DIB color tables have their colors stored BGR not RGB (so flip).
			BYTE red = paletteEntries [i].peBlue;
			BYTE green = paletteEntries [i].peGreen;
			BYTE blue = paletteEntries [i].peRed;
			packedPaletteEntries [i] = pack (red, green, blue, alpha);
		} 

		//Move the bits.
    	for	(long j = height; j > 0; j--) {
			//REVERSES: destination -= width; 
			unsigned char* to = destination; BYTE *from = source; 
			for (int i = width; i > 0; i--) { 
				*to++ = (unsigned char)packedPaletteEntries [*from++];
			}
			destination += width; 
      		source += byteWidth; 
    	} 
	} else {
		//We only handle 8 and 24 bits so far...
		log ("\nRead BMP heightmap can only handle 8 and 24 bit textures, not %d.", pixelSize); 
		delete [] m_pHeights; DeleteObject (bitmapHandle); return false;
	}

    DeleteObject (bitmapHandle); return true;
}

bool HeightMap::readTGAHeightMap (char *fileName) {
	//Creates either an RGBA texture or an RGB texture depending on whether or not the file
	//contains alpha bits... 24 bit TGA textures have not been tested...

	FILE *file = fopen (fileName, "rb");
	if (file == NULL) {::log ("\nUnable to open texture %s", fileName); return NULL;}

	#define logError(message) {log (message, fileName); fclose (file); return NULL;}
	struct TGAHeader {
		byte idLength, colorMapType, imageType, colorMapSpecification [5];
		short xOrigin, yOrigin, imageWidth, imageHeight;
		byte pixelDepth, imageDescriptor;
	};

	TGAHeader header;
	if (fread (&header, 1, sizeof (header), file) != sizeof (header))
		logError ("\nTGA file \"%s\" appears to be truncated.");
	if (header.colorMapType != 0) //1=>has color map, 0=>does not have color map.
		logError ("\nCan't read \"%s\" since it's paletted.");
	if (header.imageType != 2) //0..11; 2=>uncompressed true-color
		logError ("\nCan't read \"%s\" since it's compressed or not true color.");
	if (header.pixelDepth != 32 && header.pixelDepth != 24) {
		log ("\nFile \"%s\" is a %d bit .TGA file, need 24 or 32.", fileName, header.pixelDepth); 
		fclose (file); return NULL;
	}

	bool hasAlpha = header.pixelDepth == 32;

	//Allocate space for the heightmap.
	const long width = header.imageWidth; const long height = header.imageHeight; 
	m_pHeights = new unsigned char[width * height];

	// Set the heightmap dimensions
	m_length = height;
	m_width = width;

	if (m_pHeights == NULL) logError ("\nFailed to create heightmap for \"%s\".");

	//Create a texture to store the data in.
	long numberOfPixels = width * height;
	long bytesSize = numberOfPixels * (hasAlpha ? 4 : 3);
	BYTE *localBytes = new BYTE [bytesSize]; BYTE *destination = (BYTE *) localBytes;
 
	long bytesRead = (long)fread (localBytes, sizeof (BYTE), bytesSize, file);
	if (bytesRead == 0) {delete [] m_pHeights; logError ("\nUnable to read all the bytes in file \"%s\".");}
	fclose (file); //From here on, no longer have "close the file" as a pending action...

	//TGA is stored as BGR(A). Swizzle bits into RGB(A) format
	if (hasAlpha) {
		struct _RGBA {BYTE r, g, b, a;}; _RGBA *pixel = (_RGBA*) localBytes;
		for (long i = 0; i < numberOfPixels; i++, pixel++) {
			m_pHeights[i] = (unsigned char)pixel->r;
			BYTE oldR = pixel->r; pixel->r = pixel->b; pixel->b = oldR;
		}
	} else {
		struct _RGB {BYTE r, g, b;}; _RGB *pixel = (_RGB*) localBytes;
		for (long i = 0; i < numberOfPixels; i++, pixel++) {
			m_pHeights[i] = (unsigned char)pixel->r;
			BYTE oldR = pixel->r; pixel->r = pixel->b; pixel->b = oldR;
		}
	}

    return true;
	#undef logError
}