#ifndef MACROS_H
#define MACROS_H

//Macro "DECLARECOLLECTION" below declares a type of vector, e.g., DECLARECOLLECTION (Face);
//creates a vector of Face pointers.

//There are two ways to use such a collection

//	1. Typical way of using a collection:
//
//			FaceCollection faces; 
//			Face *face1 = ...; faces.push_back (face1); 
//			Face *face2 = ...; faces.push_back (face2);
//				...
//			LOOPVECTOR (index, faces)
//				... faces [index] .... //provides an element of type Face *.
//			ENDLOOP
//			deleteFaceCollectionEntries (faces); //faces is deleted when the object containing it is deleted...

//	2. More unusual way (using a pointer to a collection):
//
//			FaceCollection *faces = new FaceCollection; 
//			Face *face1 = ...; faces->push_back (face1); 
//			Face *face2 = ...; faces->push_back (face2);
//				...
//			LOOPVECTOR (index, *faces)
//				... (*faces) [index] .... //provides an element of type Face *.
//			ENDLOOP
//			deleteFaceCollection (faces); //deletes the entries and also faces itself...

#define DECLARECOLLECTION(type) \
	typedef vector <type *> type##Collection; \
	typedef vector <type *>::iterator type##CollectionIterator; \
		\
	inline void delete##type##CollectionEntries (type##Collection &collection) { \
		for (unsigned long index = 0; index < collection.size (); index++) { \
			delete collection [index]; \
		} \
	} \
	inline void delete##type##Collection (type##Collection *collection) { \
		delete##type##CollectionEntries (*collection); \
		delete collection; \
	}

#define LOOPVECTOR(variable,vector) {for (unsigned long variable = 0; variable < (vector).size (); variable++) {
#define ENDLOOP }}

//Macro "DECLAREDICTIONARY" below declares a type of map, e.g., DECLAREDICTIONARY (Face)
//creates a map that associates strings with Face pointers.

//There are two ways to use such a collection

//	1. Typical way of using a dictionary:
//
//			FaceDictionary faces; 
//			Face *face1 = ...; put (faces, "name1", face1); //Storing
//			Face *face2 = ...; put (faces, "name2", face2); 
//				...
//			Face *face = get (faces, "name1"); //Retrieving: get 0 if not found...
//			LOOPDICTIONARY (face, faces, Face)
//				... face .... //face is locally declared by the loop macro to be type Face * and initialized...
//			ENDLOOP
//			deleteFaceDictionaryEntries (faces); //faces is deleted when the object containing it is deleted...

//	2. More unusual way (using a pointer to a collection):
//
//			FaceDictionary *faces = new FaceDictionary; 
//			As above but use "faces->" instead of "faces." or *faces instead of faces.
//			deleteFaceDictionary (faces); //deletes the entries and also faces itself...

#define LOOPDICTIONARY(variable,map,type) \
	{for (type##DictionaryIterator iterator = (map).begin (); iterator != (map).end (); ++iterator) { \
		type *variable = iterator->second;

#define DECLAREDICTIONARY(type) \
	typedef map <string, type *> type##Dictionary; \
	typedef type##Dictionary::iterator type##DictionaryIterator; \
		\
	inline void delete##type##DictionaryEntries (type##Dictionary &collection) { \
		LOOPDICTIONARY (key, collection, type) \
			delete key; \
		ENDLOOP \
	} \
	inline void delete##type##Dictionary (type##Dictionary *collection) { \
		delete##type##DictionaryEntries (*collection); \
		delete collection; \
	} \
		\
	inline type *get (type##Dictionary &collection, char *key) { \
		type##DictionaryIterator result = collection.find (key); \
		if (result == collection.end ()) return 0; /*Key not there.*/ \
		return result->second; \
	} \
	inline void put (type##Dictionary &collection, char *key, type *value) { \
		collection.insert (type##Dictionary::value_type (key, value)); \
	}

#define INSTANCEOF(testClass, object) (dynamic_cast<testClass*>(object) != NULL)

#endif