//*****************************************************************************************//
//                                Point/Vector (synonyms)                                  //
//*****************************************************************************************//

#ifndef pointsModule
#define pointsModule 

inline float squared (float input) {return input * input;}
inline float asRadians (float degrees) {return degrees * (3.14159 / 180.0);}
inline float asDegrees (float radians) {return radians * (180.0 / 3.14159);}
inline float absolute (float number) {return fabs (number);} //Let the compiler pick
inline long absolute (long number) {return abs (number);} //the right absolute value function.

inline float epsilon () {return 1.0e-5;};
inline float squaredEpsilon () {return 1.0e-10;};
inline float inverseEpsilon () {return 1.0 / epsilon ();}

typedef struct Point Vector; //Make Vector and Point synonyms.
class Transformation; //Forward reference.
class BinaryFileStreamWriter;
class BinaryFileStreamReader;

struct Point {
	float x, y, z;
	
	inline Point *newCopy () {Point *copy = new Point; copy->x = x; copy->y = y; copy->z = z; return copy;}
	inline void zero () {x = y = z = 0.0;}
	inline bool	operator != (const Vector &p) const {return ((x!=p.x) || (y!=p.y) || (z!=p.z));} 
	inline bool	operator == (const Vector &p) const {return ((x==p.x) && (y==p.y) && (z==p.z));} 
	inline Vector operator * (const float &d) const {Vector result = {x*d, y*d, z*d}; return result;}
    inline void operator *= (const float &d) {x*=d; y*=d; z*=d;}
	inline Vector operator / (const float &d) const {Vector result = {x/d, y/d, z/d}; return result;}
    inline void operator /= (const float &d) {x/=d; y/=d; z/=d;}
	inline Vector operator + (const Vector &p) const {Vector result = {x+p.x, y+p.y, z+p.z}; return result;}
    inline void operator += (const Vector &p) {x+=p.x; y+=p.y; z+=p.z;}
    inline void operator += (const float &d) {x+=d; y+=d; z+=d;}
	inline Vector operator - (const Vector &p) const {Vector result = {x-p.x, y-p.y, z-p.z}; return result;}
    inline void operator -= (const Vector &p) {x-=p.x; y-=p.y; z-=p.z;}
    inline void operator -= (const float &d) {x-=d; y-=d; z-=d;}
	inline Vector operator - () const {Point result = {-x, -y, -z}; return result;}; //unary
	inline float squaredLength () {return x*x + y*y + z*z;}
	inline float length () {return sqrt (squaredLength ());}
	inline float dot (Vector &p) {return x*p.x + y*p.y + z*p.z;}

	inline Vector cross (const Vector &v) const {
		//Vector cross product (this cross v).
		Point result = {y*v.z-z*v.y, z*v.x-x*v.z, x*v.y-y*v.x};
		return result;
	}

	inline Point pointCross (Point &p2, Point &p3) {
		//Point cross product assuming p1,p2,p3 are counter-clockwise.
		//Create vectors from p1 (this) to p2, and p1 to p3.
		Vector vector1 = {p2.x - x, p2.y - y, p2.z - z};
		Vector vector2 = {p3.x - x, p3.y - y, p3.z - z};
		return vector1.cross (vector2);
	}
	
	inline void normalize () {//destructively
		float length = sqrt (squaredLength ());
		if (absolute (length) > epsilon ()) {
			float oneOverLength = 1.0 / length;
			x *= oneOverLength; y *= oneOverLength; z *= oneOverLength;
		}
	}

	inline float squaredDistanceTo (Point &point) {
		Point difference = point - *this;
		return difference.dot (difference);
	}

	Point operator * (const Transformation &transformation) const;

	inline static Point axisX() {Point p = {1.0f, 0.0f, 0.0f}; return p;}
	inline static Point axisY() {Point p = {0.0f, 1.0f, 0.0f}; return p;}
	inline static Point axisZ() {Point p = {0.0f, 0.0f, 1.0f}; return p;}

	bool serialize(BinaryFileStreamWriter *output) {return false;}
	bool deserialize(BinaryFileStreamReader *input);
};

DECLARECOLLECTION (Point);

#endif
