#include "includes.h"

void Terrain::initialize()
{

	// @pre - The properties of the Terrain must have been set at this point

	// Create the heightmap object
	m_pHeightMap = new HeightMap();

	if(m_pHeightMap)
	{
		if(m_pHeightMap->loadHeightMap((char *)m_heightMapFile.c_str()))
		{
			// Create the terrain vertice and indice data
			//------------
			int	width	= m_pHeightMap->getWidth();
			int length	= m_pHeightMap->getLength();

			// Calculate the grid size to fill the dimensions
			float fwidth	= width;
			float flength	= length;
			// There is a grid of NxM squares where N and M are one less than the number of vertices on each side
			m_gridSizeX		= m_sizeX / (fwidth - 1.0);
			m_gridSizeZ		= m_sizeZ / (flength - 1.0);

			// 1:1 vertices per heightmap value
			m_pVertices		= (TERRAINPOINT*)malloc(sizeof(TERRAINPOINT)*width*length);

			// 1:1 texture coords per heightmap value
			m_pTexCoords	= (TERRAINTEX*)malloc(sizeof(TERRAINTEX)*length*width);

			// The heightmap data is stored in (length - 1) triangle strips with (2 * width) indices per strip. (top and bottom indices)

			// e.g. Width = 3, Length = 2
			//  1    3    5
			//  |   /|   /|
 			//  |  / |  / |
			//  | /  | /  |
			//  |/   |/   |
			//  2    4    6
			//
			// #indices = 6
			// and (2 - 1) = 1 triangle strip

			m_pIndices = (unsigned int*)malloc(sizeof(unsigned int) * ( (length - 1) * (width * 2) ) );
			
			// Create the indices
			//-----------

			// Loop through the rows building a triangle strip for each row
			for(int j=0; j<(length - 1); j++)
			{
				// Loop through the columns adding to the triangle strip
				for(int i=0; i < (2 * width); i+=2)
				{
					// Reverse the order of these two indices to get Clock Wise starting triangle strips
					m_pIndices[j * (2 * width) + i] = (j + 1) * width + (i/2);
					m_pIndices[j * (2 * width) + i + 1] = j * width + (i/2);
				}
			}

			// Load the heightmap values into the y component of the vertices
			float h;
			for(int j=0; j<length; j++)
			{
				for(int i=0; i<width; i++)
				{
					// The point is mapped by:
					// i => x
					// j => -z
					m_pVertices[j*width + i].x = i * m_gridSizeX;
					m_pVertices[j*width + i].z = -j * m_gridSizeZ;

					// Map the height. Convert the height from unsigned char to float from 0->1 first
					h = (float)(*m_pHeightMap->getHeight(i, j));
					m_pVertices[j*width + i].y = m_height * ( h / 255.0f );
				}
			}

			//Create the texture coordinates
			for(int j=0; j<length; j++)
			{
				for(int i=0; i<width; i++)
				{
					m_pTexCoords[j*width + i].tx	= ((float)i) / fwidth;
					m_pTexCoords[j*width + i].ty	= ((float)j) / flength;
				}
			}

			if(GLExtensions::get()->vertexBuffersAvailable())
			{

				// Create the vertex buffer and copy the content into it
				GLExtensions::get()->glGenBuffersARB(1, &m_vertexBuffer);
				GLExtensions::get()->glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_vertexBuffer);
				GLExtensions::get()->glBufferDataARB(GL_ARRAY_BUFFER_ARB, sizeof(TERRAINPOINT)*width*length, m_pVertices, GL_STATIC_DRAW_ARB);

				// Create the index buffer and copy the content into it
				GLExtensions::get()->glGenBuffersARB(1, &m_indexBuffer);
				GLExtensions::get()->glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_indexBuffer);
				GLExtensions::get()->glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, sizeof(unsigned int)*(2*width)*(length-1), m_pIndices, GL_STATIC_DRAW_ARB);

				// Create the texcoords buffer and copy the content into it
				GLExtensions::get()->glGenBuffersARB(1, &m_textureBuffer);
				GLExtensions::get()->glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_textureBuffer);
				GLExtensions::get()->glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, sizeof(TERRAINTEX)*width*length, m_pTexCoords, GL_STATIC_DRAW_ARB);
			}
		}
	}
}

Terrain::~Terrain()
{

	// Delete the vertex buffer objects if they were allocated 
	if( GLExtensions::get()->vertexBuffersAvailable())
	{
		if( m_vertexBuffer )
			GLExtensions::get()->glDeleteBuffersARB(1, &m_vertexBuffer);
		if( m_indexBuffer )
			GLExtensions::get()->glDeleteBuffersARB(1, &m_indexBuffer);
		if( m_textureBuffer )
			GLExtensions::get()->glDeleteProgramsARB(1, &m_textureBuffer);
	}

	if(m_pVertices)
	{
		free(m_pVertices);
		m_pVertices = NULL;
	}

	if(m_pTexCoords)
	{
		free(m_pTexCoords);
		m_pTexCoords = NULL;
	}

	if(m_pIndices)
	{
		free(m_pIndices);
		m_pIndices = NULL;
	}

	if(m_pHeightMap)
	{
		delete m_pHeightMap;
		m_pHeightMap = NULL;
	}

}

void Terrain::draw()
{
	Transformation matrix;
	matrix.setToIdentity();
		
	matrix.translateBy(m_position);
	matrix.rotateBy(m_rotation);
		
	glPushMatrix();
	glMultMatrixd(matrix);

	// Centre the terrain at the position
	int length = m_pHeightMap->getLength();
	int width = m_pHeightMap->getWidth();
	glTranslated(-m_sizeX*0.5f, 0.0, 0.5f*m_sizeZ);
	
	// Activate the texture
	TextureList::Get()->GetTexture(m_colorMap)->activate();

	//Check if we are drawing with Vertex buffers
	bool useVBOs = GLExtensions::get()->vertexBuffersAvailable();

	glEnableClientState( GL_VERTEX_ARRAY );
	glEnableClientState( GL_TEXTURE_COORD_ARRAY );

	// Activate the buffers if we are using them to draw
	if( useVBOs )
	{
		GLExtensions::get()->glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_vertexBuffer);
		glVertexPointer(3, GL_FLOAT, 0, 0);
		GLExtensions::get()->glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_textureBuffer);
		glTexCoordPointer(2, GL_FLOAT, 0, 0);
		GLExtensions::get()->glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_indexBuffer);
	}
	// Otherwise use the system memory copies
	else
	{
		glVertexPointer( 3, GL_FLOAT, 0, m_pVertices );
		glTexCoordPointer( 2, GL_FLOAT, 0, m_pTexCoords );
	}

	// Loop through the triangle strips
	for(int j=0; j<(length-1); j++)
	{
		//glColor3f (0.0f + (((float)j)/((float)(length-1))), 0.0f, 0.0f);
		if(useVBOs)
		{
			// Draw each row of the terrain
			glDrawElements(GL_TRIANGLE_STRIP, width*2, GL_UNSIGNED_INT, (const GLvoid*)(sizeof(unsigned int) * j * (2 * width)));
		}
		else
		{
			// Draw each row of the terrain
			glDrawElements(GL_TRIANGLE_STRIP, width*2, GL_UNSIGNED_INT, &(m_pIndices[j * (2 * width)]));
		}
	}

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisable (GL_TEXTURE_2D);

	/* FOR DEBUG PURPOSES
	for(int j=0; j<length-1; j++)
	{
		for(int i=0; i<(2*width); i++)
		{
			unsigned int a = m_pIndices[j*(width*2) + i];
			TERRAINPOINT* p = &m_pVertices[a];

			glPushMatrix();
			glFrontFace (GL_CW);
			glColor3f (0.03f, 0.53f, 0.96f);
			glTranslated(p->x, p->y, p->z);
			glutSolidSphere (0.5, 3, 3);
			glFrontFace (GL_CCW);
			glPopMatrix();
		}
	}*/


	glPopMatrix();

}

void Terrain::processProperties() {
	char buffer[32];
	PropertyMapIterator iterator = m_properties.find("heightMap");
	if (iterator != m_properties.end() ) {		
		m_heightMapFile = iterator->second;
	}

	iterator = m_properties.find("colorMap");
	if(iterator != m_properties.end() )
	{
		strcpy(buffer, iterator->second);
		sscanf(buffer, "%d", &m_colorMap);
	}

	iterator = m_properties.find("heightRange");
	if(iterator != m_properties.end() )
	{
		strcpy(buffer, iterator->second);
		sscanf(buffer, "%f", &m_height);
	}

	iterator = m_properties.find("width");
	if(iterator != m_properties.end() )
	{
		strcpy(buffer, iterator->second);
		sscanf(buffer, "%f", &m_sizeX);
	}

	iterator = m_properties.find("length");
	if(iterator != m_properties.end() )
	{
		strcpy(buffer, iterator->second);
		sscanf(buffer, "%f", &m_sizeZ);
	}

}
