#pragma once

// Declare the event types
// This is not an enum because I wish to allow other classes
// to add eventTypes without having to redeclare the enum
#define	INVALID				-1
#define	REDISPLAY			1
#define	RESIZE				2	
#define	MOUSEMOVE			3
#define	MOUSEUP				4
#define	MOUSEDOWN			5
#define	KEYUP				6
#define	KEYDOWN				7
#define REDISPLAYREQUEST	8
#define	SCROLLVALUE			9
#define BUTTONPRESS			10

typedef int WINDOWID;

typedef int EVENTTYPE;

enum EVENTDIRECTION
{
	UP,
	DOWN,
	BOTH
};

class VDKey;

// Generic base class event
class VDEvent
{
protected:
	// The direction the message is passed in the window heirarchy
	EVENTDIRECTION	m_direction;
	EVENTTYPE		m_type;

	// Override this method in the base class that derives from VDEvent
	// Ensure the type is set to a valid type in this method
	// Call this method in the constructor of the derived class.
	// I Couldn't put this method into the constructor of VDEvent.
	virtual void setType() = 0;

public:

	// By default events waterfall down the window heirarchy
	VDEvent(void);
	virtual ~VDEvent(void) {}

	virtual void			setDirection(EVENTDIRECTION d)	{ m_direction = d; }

	virtual EVENTDIRECTION	getDirection()					{ return m_direction; }
	virtual EVENTTYPE		getType()						{ return m_type; }

};

class VDWindowEvent : public VDEvent
{
protected:
	WINDOWID	m_id;
public:
	VDWindowEvent(WINDOWID id)	{ m_id = id; }
	virtual ~VDWindowEvent()	{}

	int getWindowId()	{ return m_id; }
};

// Button press output event
class VDButtonPressEvent : public VDWindowEvent
{
protected:
	virtual void setType()	{ m_type = BUTTONPRESS; }
public:
	VDButtonPressEvent(WINDOWID id) : VDWindowEvent(id) { setType(); m_direction = UP;}
	virtual ~VDButtonPressEvent() {}
};

// Scroll bar output event
class VDScrollValueEvent : public VDWindowEvent
{
protected:
	int m_scrollValue;

	virtual void setType() { m_type = SCROLLVALUE; }

public:
	VDScrollValueEvent(WINDOWID id, int scrollValue) : VDWindowEvent(id) { m_scrollValue = scrollValue; setType(); m_direction = UP; }
	virtual ~VDScrollValueEvent() {};

	int	getScrollValue() { return m_scrollValue; }
};

// Redisplay Event
class VDRedisplayEvent : public VDEvent
{
protected:
	virtual void setType() { m_type = REDISPLAY; }

public:

	VDRedisplayEvent(void) { setType(); }
	~VDRedisplayEvent(void) {}

};

// Resize Event
class VDResizeEvent : public VDEvent
{
protected:
	int m_width;
	int m_height;

	virtual void setType() { m_type = RESIZE; }

public:

	VDResizeEvent(int width, int height) 
	{ 
		m_width		= width;
		m_height	= height;
		setType(); 
	}
	~VDResizeEvent(void) {}

	int getNewWidth() { return m_width; }
	int getNewHeight() { return m_height; }

};

// base Mouse event
class VDMouseEvent : public VDEvent
{
protected:
	int m_x;
	int m_y;
public:
	VDMouseEvent(int x, int y) { m_x = x; m_y = y; }
	virtual ~VDMouseEvent()		{}

	int	getMouseX() { return m_x; }
	int	getMouseY() { return m_y; }
};

// Mouse Move Event
class VDMouseMoveEvent : public VDMouseEvent
{
protected:
	virtual void setType() { m_type = MOUSEMOVE; }
public:
	VDMouseMoveEvent(int x, int y) : VDMouseEvent(x, y) { setType(); }
	virtual ~VDMouseMoveEvent(void) {};
};

// Mouse button event
class VDMouseButtonEvent : public VDMouseEvent
{
protected:
	// The button pressed or lifted 
	int m_button;

public:
	VDMouseButtonEvent( int button, int x, int y ) : VDMouseEvent(x, y) { m_button = button; }
	virtual ~VDMouseButtonEvent(void) {};

	int	getButton() { return m_button; }
};

// MOUSEUP event
class VDMouseUpEvent : public VDMouseButtonEvent
{
protected:
	virtual void setType() { m_type = MOUSEUP; }

public:
	VDMouseUpEvent( int button, int x, int y) : VDMouseButtonEvent(button, x, y) { setType(); }
	virtual ~VDMouseUpEvent() {};
};

// MOUSEDOWN event
class VDMouseDownEvent : public VDMouseButtonEvent
{
protected:
	virtual void setType() { m_type = MOUSEDOWN; }

public:
	VDMouseDownEvent( int button, int x, int y) : VDMouseButtonEvent(button, x, y) { setType(); }
	virtual ~VDMouseDownEvent() {};
};

// Key event
class VDKeyEvent : public VDEvent
{
protected:
	VDKey	m_key;
public:
	VDKeyEvent(VDKey key) { m_key = key; }
	virtual ~VDKeyEvent() {};

	VDKey	getKey() { return m_key; }
};

// Key Down event
class VDKeyDownEvent : public VDKeyEvent
{
protected:
	virtual void setType() { m_type = KEYDOWN; }
public:
	VDKeyDownEvent(VDKey key) : VDKeyEvent(key) { setType(); }
	virtual ~VDKeyDownEvent() {};
};

// Key Up event
class VDKeyUpEvent : public VDKeyEvent
{
protected:
	virtual void setType() { m_type = KEYUP; }
public:
	VDKeyUpEvent(VDKey key) : VDKeyEvent(key) { setType(); }
	virtual ~VDKeyUpEvent() {};
};

// Request redraw event
class VDRedisplayRequestEvent : public VDEvent
{
protected:
	virtual void setType() { m_type = REDISPLAYREQUEST; }
public:
	VDRedisplayRequestEvent(void) { setType(); m_direction = UP; }
	virtual ~VDRedisplayRequestEvent(void) {}
};