#include "vd.h"

/*static*/ VDMainWindow* VDMainWindow::s_pMainWindow = NULL;

VDMainWindow::VDMainWindow(int gridX, int gridY)
{
	// Set the id of the one and only main window
	m_id = MAINWINDOW;
	m_gridSizeX = gridX;
	m_gridSizeY = gridY;

	// Set to the resolution initially
	m_bounds.setBounds(0, 0, m_gridSizeX, m_gridSizeY);

	// Reset the message box
	m_pMessageBox = NULL;

	// reset the custom window
	m_pCustomWindow = NULL;

	setupControls();
}

VDMainWindow::~VDMainWindow(void)
{
	// Destroy the message box if we exit and its still up
	if(m_pMessageBox)
	{
		delete m_pMessageBox;
		m_pMessageBox = NULL;
	}

	VDConsole::destroy();
}

// Create the menus
void VDMainWindow::setupControls()
{

	// Create a console
	VDConsole::create();
	VDConsole* pConsole = VDConsole::get();	
	VDRectangleContainer<float> bounds;
	bounds.setBounds(10, 50, 790, 550);
	pConsole->resizeControls(bounds);
	pConsole->setupControls();
	// Don't add the console to the child windows.. treat it specially
	pConsole->setParentWindow(this);

}

bool VDMainWindow::processEvent(VDEvent *pEvent)
{

	if( pEvent->getType() == REDISPLAY )
	{

		// Redraw the window
		//************
		VDTheme* t = VDThemeStore::get()->getCurrentTheme();
		if( t == NULL )
		{
			::log("Unable to grab current theme from ThemeStore");
			return true;
		}
		
		t->drawMainWindowBackground();
		t->beginThemeDrawing(m_gridSizeX, m_gridSizeY);

		// Draw the custom window
		if( m_pCustomWindow)
			m_pCustomWindow->processEvent(pEvent);

		// if the console is visible draw that
		VDConsole::get()->processEvent(pEvent);

		// Draw the message box
		if(m_pMessageBox)
		{
			m_pMessageBox->processEvent(pEvent);
		}

		t->endThemeDrawing();

		glutSwapBuffers ();	
	}
	else if( pEvent->getType() == RESIZE )
	{
		VDResizeEvent* pResizeEvent = dynamic_cast<VDResizeEvent*>(pEvent);

		// Don't handle this event if it doesn't cast to VDResizeEvent
		if(!pResizeEvent)
			return true;

		// Recalculate gridSize
		m_gridSizeX = pResizeEvent->getNewWidth();
		m_gridSizeY = pResizeEvent->getNewHeight();

		VDRectangleContainer<float> bounds;
		bounds.setBounds(0, 0, pResizeEvent->getNewWidth(), pResizeEvent->getNewHeight());
		resizeControls(bounds);
		if( m_pMessageBox )
			m_pMessageBox->resizeControls(bounds);

		//************
		glViewport(0, 0, (GLsizei)pResizeEvent->getNewWidth(), (GLsizei)pResizeEvent->getNewHeight());
	}
	else if( pEvent->getType() == REDISPLAYREQUEST )
	{
		// Post a redisplay since a child window has requested it
		glutPostRedisplay();
	}
	else if( pEvent->getType() == MOUSEMOVE || pEvent->getType() == MOUSEDOWN || pEvent->getType() == MOUSEUP )
	{
		// Convert the mouse coordinates to main window coordinates
		VDMouseEvent* pMouseEvent = dynamic_cast<VDMouseEvent*>(pEvent);
		if(pMouseEvent == NULL)
			return false;

		int x = pMouseEvent->getMouseX();
		int y = pMouseEvent->getMouseY();
		
		// grid coordinates go from bottom left (0, 0) to top right (gridX, gridY)
		// y is inverted and scaled
		// x is just scaled

		x = (x / m_bounds.getWidth()) * m_gridSizeX;
		y = ( 1 - (y / m_bounds.getHeight())) * m_gridSizeY;

		// New event to send
		VDMouseEvent* e = NULL;
		if( pEvent->getType() == MOUSEMOVE )
		{
			e = new VDMouseMoveEvent(x, y);
		}
		else
		{
			VDMouseButtonEvent * pMouseButton = dynamic_cast<VDMouseButtonEvent*>(pEvent);
			if( !pMouseButton )
			{
				::log("Main window encountered an invalid mouse button event");
				return false;
			}
			if( pEvent->getType() == MOUSEDOWN )
			{
				e = new VDMouseDownEvent(pMouseButton->getButton(), x, y);
			}
			else
			{
				e = new VDMouseUpEvent(pMouseButton->getButton(), x, y);
			}			
		}

		return sendEvent(e);

	}
	else
	{
		if( pEvent->getType() == KEYDOWN )
		{
			VDKeyDownEvent* e = dynamic_cast<VDKeyDownEvent*>(pEvent);
			if ( e == NULL )
			{
				::log("Invalid key down object associated with keydown type");
				return false;
			}

			// Message box gobbles all
			if(m_pMessageBox)
			{
				return m_pMessageBox->processEvent(e);
			}

			VDKey k = e->getKey();
			if( k.getKey() == VD_TILDE )
			{
				if( VDConsole::get()->isVisible() )
					VDConsole::get()->hide();
				else
					VDConsole::get()->show();
				return true;
			}
			else if( k.getKey() == VD_ESCAPE )
			{
				VDExitCommand* e = new VDExitCommand();
				e->execute();
				// Post a redisplay to show the message box
				glutPostRedisplay();
				return true;
			}
		}

		return sendEvent(pEvent);
	}

	return false;
}

void VDMainWindow::resizeControls(VDRectangleContainer<float> bounds)
{
	// Resize the main window only
	m_bounds = bounds;

	// Resize any controls based on the size of the window
}

void VDMainWindow::messageBox(MESSAGEBOXTYPE type, string message, VDMessageBoxCallBack* pCallBack)
{
	// Save the callback object
	m_pMessageBoxCallBack = pCallBack;

	// Set the dialog to tell us when the value has been selected
	// so the main window can delete the window.
	// Then pass the message on to the callback object
	m_pMessageBox = new VDMessageBox(type, message, this);
	VDRectangleContainer<float> bounds;
	bounds.setBounds((m_bounds.getWidth()/2) - 200, (m_bounds.getHeight()/2) - 40, (m_bounds.getWidth()/2) + 200, (m_bounds.getHeight()/2) + 40);
	m_pMessageBox->resizeControls(bounds);
	m_pMessageBox->setupControls();
	// Get the display messages from the message box
	m_pMessageBox->setParentWindow(this);
}

void VDMainWindow::messageBoxResponse(MESSAGEBOXRESPONSE r)
{
	// Notify the interested object
	if(m_pMessageBoxCallBack)
	{
		m_pMessageBoxCallBack->messageBoxResponse(r);
	}

	// Erase our message box pointer. The message box will delete itself via the garbage can
	m_pMessageBox = NULL;
	glutPostRedisplay();
	
}

bool VDMainWindow::sendEvent(VDEvent* pEvent)
{

	// If the message box is there it gets first dibs
	if(m_pMessageBox)
		return m_pMessageBox->processEvent(pEvent);

	// Next check if the console is visible and gobbles the event
	if( VDConsole::get()->isVisible() )
		return VDConsole::get()->processEvent(pEvent);
	
	// Finally pass it along the application window tree
	if( m_pCustomWindow)
		return m_pCustomWindow->processEvent(pEvent);

	return false;
}
